import types

from selenium.webdriver.common.options import ArgOptions
from selenium.webdriver import ChromeOptions, FirefoxOptions, EdgeOptions, IeOptions, WebKitGTKOptions
from selenium.webdriver import Chrome, Firefox, Edge, Ie, WebKitGTK, Safari
from selenium.webdriver.safari.options import Options as SafariOptions
from .utils import Config, BrowserType
from .action_executor import ActionExecutor



class Browser:
  EQUVALENCES = equivalences = {
    BrowserType.CHROME.value: [ChromeOptions, Chrome],
    BrowserType.FIREFOX.value: [FirefoxOptions, Firefox],
    BrowserType.EDGE.value: [EdgeOptions, Edge],
    BrowserType.IE.value: [IeOptions, Ie],
    BrowserType.SAFARI.value: [SafariOptions, WebKitGTK],
    BrowserType.WEBKITGTK.value: [WebKitGTKOptions, WebKitGTK]
  }


  def __init__(self, config: dict) -> None:
    self._config = Config(config)
    self._load_browser_props()
    self._action_executor = ActionExecutor(self._driver, self._config)

  @property
  def config(self) -> Config:
    return self._config


  def _load_browser_props(self):
    """Loads browser driver instance and options"""
    if self._config.get('general.browser.type') not in self.EQUVALENCES:
      raise Exception('Browser type not supported')
    else:
      options_class, driver_class = self.EQUVALENCES[self._config.get('general.browser.type')]
      self._options = self._create_options(options_class)
      self._driver = driver_class(self._config.get('general.browser.driver'), options=self._options)


  def _create_options(self, options_class):
    options = options_class()

    for opt_key, opt_value in self._config.get('general.browser.options', {}).items():
      if hasattr(options_class, opt_key):
        if callable(getattr(options_class, opt_key)):
          self._load_browser_method(options, opt_key, opt_value)
        else:
          setattr(options, opt_key, opt_value)

    return options


  def _load_browser_method(self, options, opt_key, opt_value) -> None:
    if isinstance(opt_value, list):
      for value in opt_value:
        if "key" in value and "activate" in value and value["activate"]:
          getattr(options, opt_key)("--{}".format(value["key"]))
        elif "key" in value and "value" in value:
          getattr(options, opt_key)(value["key"], value["value"])
        else:
          pass
    else:
      getattr(options, opt_key)(**opt_value)


  def _load_browser_options(self) -> None:
    equivalences = {
      BrowserType.CHROME.value: ChromeOptions,
      BrowserType.FIREFOX.value: FirefoxOptions,
      BrowserType.EDGE.value: EdgeOptions,
      BrowserType.IE.value: IeOptions,
      BrowserType.SAFARI.value: SafariOptions,
      BrowserType.WEBKITGTK.value: WebKitGTKOptions
    }

    if self._config.get('general.browser.type') not in equivalences:
      raise Exception('Browser type not supported')
    else:
      self._options = equivalences[self._config.get('general.browser.type')]()

      for opt_key, opt_value in self._config.get('general.browser.options', {}).items():
        if hasattr(self._options, opt_key):
          if callable(getattr(self._options, opt_key)):
            self._load_browser_method(opt_key, opt_value)
          else:
            self._load_browser_attribute(opt_key, opt_value)


  def run(self) -> None:
    self._action_executor.run()


  @property
  def driver(self) -> ArgOptions | Chrome | Firefox | Edge | Ie | WebKitGTK | Safari:
    return self._driver



import unittest
import selenium_loader
import tests.common as common

from unittest.mock import Mock, mock_open
from selenium.webdriver.remote.webdriver import WebDriver



class ActionExecutorTest(unittest.TestCase):
  def setUp(self):
    self._driver_mock = Mock(spec=WebDriver)
    self._driver_mock.__enter__ = lambda self: self
    self._driver_mock.__exit__ = lambda self, exc_type, exc_val, exc_tb: None

    self._config_mock = Mock(spec=selenium_loader.Config)


  def test_call_method_dynamically_with_none_object(self):
    try:
      selenium_loader.ActionExecutor.call_method_dynamically(None, "test")
    except Exception as e:
      assert str(e) == "Imposible call method on None object"


  def test_call_method_dynamically_with_object_has_not_method(self):
    test_object = common.TestObject()

    try:
      selenium_loader.ActionExecutor.call_method_dynamically(test_object, "test")
    except Exception as e:
      assert str(e) == "Object {} has no method {}".format(test_object.__class__.__name__, "test")


  def test_call_method_dynamically_with_object_with_different_kind_of_method_and_parameters(self):
    test_object = common.TestObject()
    selenium_loader.ActionExecutor.call_method_dynamically(test_object, "with_only_one_param", 1)
    selenium_loader.ActionExecutor.call_method_dynamically(test_object, "with_bool_param")
    selenium_loader.ActionExecutor.call_method_dynamically(test_object, "with_list_param", [1, 2])
    selenium_loader.ActionExecutor.call_method_dynamically(test_object, "with_object_param", {"param1": 1, "param2": 2, "param3": 3})

    assert test_object.one_param == 1
    assert test_object.bool_param == 1
    assert test_object.list_param == 1
    assert test_object.dict_param == 1


  def test_run_with_no_steps(self):
    self._config_mock.get.return_value = []

    action_executor = selenium_loader.ActionExecutor(self._driver_mock, self._config_mock)
    action_executor.run()

    self._config_mock.assert_not_called()


  def test_run_with_one_invalid_step(self):
    self._config_mock.get = Mock(args=("steps", []), return_value=[])

    action_executor = selenium_loader.ActionExecutor(self._driver_mock, self._config_mock)
    action_executor.run()


if __name__ == '__main__':
  unittest.main()



from selenium.webdriver.remote.webdriver import WebDriver
from .utils import Config
from .custom_web_driver_handler import CustomWebDriverHandler


class ActionExecutor:
  def __init__(self, driver: WebDriver, config: Config) -> None:
    self._driver = driver
    self._config = config
    self._custom_web_driver_handler = CustomWebDriverHandler()


  def run(self):
    with self._driver as driver:
      for step_config in self._config.get('steps', []):
        self._execute_step(driver, step_config)


  def _execute_step(self, driver: WebDriver, step_config: any) -> None:
    if self._config.get("general.verbose", False):
      print("Executing step: {}".format(step_config["name"]))

    for action_config in step_config["actions"]:
      result = self.call_method_dynamically(
        driver,
        action_config["command"],
        (action_config["parameters"] if "parameters" in action_config else None)
      )

      if result and "handle" in action_config:
        for handle_key, handle_value in action_config["handle"].items():
          if self._custom_web_driver_handler and hasattr(self._custom_web_driver_handler, handle_key) and callable(getattr(self._custom_web_driver_handler, handle_key)):
            local_result = self.call_method_dynamically(self._custom_web_driver_handler, handle_key, {**{"element": result}, **handle_value})
          else:
            local_result = self.call_method_dynamically(result, handle_key, handle_value)

          if "set_variable" in action_config:
            self._config.set(action_config["set_variable"], local_result)


  @classmethod
  def call_method_dynamically(cls, object: any, command: str, parameters: any = None) -> any:
    if not object:
      raise Exception("Imposible call method on None object")

    hasattribute = hasattr(object, command)

    if not hasattribute or (hasattribute and not callable(getattr(object, command))):
      raise Exception("Object {} has no method {}".format(object.__class__.__name__, command))

    params_type = type(parameters)

    if params_type in [str, int, float]:
      return getattr(object, command)(parameters)
    elif params_type == bool or params_type == type(None):
      return getattr(object, command)()
    elif params_type == list:
      return getattr(object, command)(*parameters)
    elif params_type == dict:
      return getattr(object, command)(**parameters)
    else:
      return None



import re, json

from .config_validator import ConfigValidator
from .browser_type import BrowserType



class Config:
  def __init__(self, raw_config: dict) -> None:
    ConfigValidator.validate(raw_config)
    self._raw_config = raw_config


  def variable_resolver(self, data: str | dict) -> str | dict:
    if not data:
      return None

    was_object = isinstance(data, dict) or isinstance(data, list)
    data = json.dumps(data) if was_object else data
    data = re.sub(
      r"\$\{(\w+)\}",
      lambda match: self._raw_config["general"]["variables"][match.group(1)] if match.group(1) in self._raw_config["general"]["variables"] else "",
      data
    )
    return json.loads(data) if was_object else data


  def get(self, key: str, default: any = None, variable_resolution: bool = True) -> any:
    default = default if not variable_resolution else self.variable_resolver(default)
    key = key if not variable_resolution else self.variable_resolver(key)
    buffer = self._raw_config if not variable_resolution else self.variable_resolver(self._raw_config)

    for key in key.split('.'):
      if key in buffer:
        buffer = buffer[key]
      else:
        return default

    return buffer


  def set(self, key: str, value: any, seed: any = None, calls: int = 0) -> any:
    seed = self._raw_config if calls == 0 and not seed else seed
    keys = key.split('.')
    current = keys[-1]

    if type(seed) == dict:
      if current in seed:
        seed[current] = value
      else:
        for local_key, local_value in seed.items():
          seed[local_key] = self.set('.'.join(keys[1:]), value, local_value, calls + 1)
          calls += 1
    else:
      if current:
        seed = value

    return seed


import os, re, pathlib, time

from selenium.webdriver.remote.webelement import WebElement


class CustomWebDriverHandler:
  @classmethod
  def submit_with_download(cls, element: WebElement, timeout: int, download_dir: str, actuator: str = 'submit') -> set[str]:
    getattr(element, actuator)()

    download_dir = pathlib.Path(download_dir).absolute()
    original_files = set(os.listdir(download_dir))
    to_ignore = len(re.findall(".crdownload", "|".join(original_files)))
    current_files = set()
    download_complete = False
    seconds = 0

    while not download_complete and seconds < timeout:
      time.sleep(1)
      seconds += 1
      current_files = set(os.listdir(download_dir))
      download_complete = not (len(re.findall(".crdownload", "|".join(current_files))) - to_ignore)

    return ["{}/{}".format(download_dir, file) for file in (current_files - original_files)]



class TestObject:
  def __init__(self) -> None:
    self.one_param = 0
    self.bool_param = 0
    self.list_param = 0
    self.dict_param = 0

  def with_only_one_param(self, param1):
    self.one_param += 1

  def with_bool_param(self):
    self.bool_param += 1

  def with_list_param(self, param1, param2):
    self.list_param += 1

  def with_object_param(self, param1 = None, param2 = None, param3 = None):
    if param1 and param2 and param3:
      self.dict_param += 1



"""Configuration loader and action executor for Selenium"""

__version__ = "0.2.0"

from .action_executor import ActionExecutor
from .utils import BrowserType, Config
from .custom_web_driver_handler import CustomWebDriverHandler
from .browser import Browser


class ConfigValidator:
  MESSAGE_FORMAT = "\"{}\" should contains the following keys: {}."
  STRUCTURE = {
    "general": {"browser": ["type", "driver"], "variables": None},
    "steps": {
      "*.name": None,
      "*.actions": {
        "*.command": None,
        "*.parameters": None
      }
    }
  }


  @classmethod
  def throw_handler(cls, message: str, throw_exception: bool) -> bool:
    if throw_exception:
      raise ValueError(message)
    else:
      return False


  @classmethod
  def validate(cls, config: dict, structure: dict|list = None, key: str = "config", throw_exception: bool = True) -> bool:
    structure = structure if structure else cls.STRUCTURE
    buffer_str = structure if isinstance(structure, dict) else {structure[i]: None for i in range(len(structure))}
    buffer_str = {key.replace("*.", ""): value for key, value in buffer_str.items()}

    if not config or not all([key in config for key in buffer_str.keys()]):
      return cls.throw_handler(cls.MESSAGE_FORMAT.format(key, ", ".join(buffer_str.keys())), throw_exception)
    else:
      for key_str, value_str in buffer_str.items():
        if value_str:
          if isinstance(config[key_str], list):
            i = len(config[key_str])
            while i > 0:
              i -= 1
              if not cls.validate(config[key_str][i], value_str, "{}[{}]".format(key_str, i)):
                return cls.throw_handler(cls.MESSAGE_FORMAT.format("{}[{}]".format(key_str, i), ", ".join(value_str.keys())), throw_exception)
          else:
            if not cls.validate(config[key_str], value_str, key_str):
              return cls.throw_handler(cls.MESSAGE_FORMAT.format(key_str, ", ".join(value_str.keys())), throw_exception)
        else:
          continue
      return True
